from jinja2 import Template


QUIZ_GRADING_PROMPT = Template('''You are a friendly professor whose task is to grade a single student answer to an exam question based on an answer key.

# Question

{{ question }}

# Answer key

{{ answer_key }}

# Reply format

Respond with a JSON string in the format of the example below. Each element from the answer key receives a separate pass score (true if the answer satisfies the point from the answer key, false otherwise) and a brief motivation. When the answer key contains specific terminology or phrasing, variations of this terminology or phrasing should be considered correct as well, provided that they clearly convey the same meaning. __Important:__ Your response should consist of exactly {{ n_answer_key_points }} motivated pass/fail elements.

{{ reply_format }}

Respond only with JSON. Do not include any additional text.

The student answer will be in the next message.''')

EXAM_VALIDATION_PROMPT = Template('''You are a university professor. Your task is to double-check the quality of an exam question. You are provided with the question itself and the answer key that you will use later to score student responses.
                                  
The answer key uses a specific notation:
                                  
- It consist of separate elements, formatted as a markdown list with '-' prefixes.
- A single element may specify a list of options of which several should be provided. This indicated with a number prefix, like this: "3:Name three of the following (…)"
                                  
Please consider that:
    
- A answer key is good when each of the elements can be scored separately and without ambiguity as correct or incorrect.
- An answer key is good when it is a minimal criterion. It does not need to be comprehensive. Students are not penalized for providing additional information that is not covered by the answer key.
- A concise answer key is better than an elaborate one.
- Each of the elements is scored as correct or incorrect; no partial credit is given for individual elements.
- You will be the one using the answer key to score the exam.

If the question and answer key are fine, simply say so! If the question or answer key needs to be improved, please provide a brief recommendation for how to update question or the answer key.

## Question

{{ question_text }}

## Answer key

- {{ answer_key }}
''')

QUALITATIVE_ERROR_ANALYSIS_PROMPT = Template('''You are a university professor. Your task is to check the quality of an exam question. You are provided with all student responses to this question that were considered partly or entirely incorrect. (But not with the correct responses.) You are also provided with the question itself and the answer key that was used to score the student responses.
        
Your task is to compare the incorrect student responses against the question and the answer key, and to indicate whether or not some student responses were unfairly considered incorrect as a result of mistakes or omissions in the answer key. If the answer key is fine, simply say so! If the answer key needs to be improved, please suggest a revised answer key.
                                             
The answer key uses a specific notation:
                                  
- It consist of separate elements, formatted as a markdown list with '-' prefixes.
- A single element may specify a list of options of which several should be provided. This indicated with a number prefix, like this: "3:Name three of the following (…)"
                                             
Please consider that:
    
- A answer key is good when each of the elements can be scored separately and without ambiguity as correct or incorrect.
- An answer key is good when it is a minimal criterion. It does not need to be comprehensive. Students are not penalized for providing additional information that is not covered by the answer key.
- A concise answer key is better than an elaborate one.
- Each of the elements is scored as correct or incorrect; no partial credit is given for individual elements.
- You will be the one using the answer key to score the exam.
                                             
When the answer key contains specific terminology or phrasing, variations of this terminology or phrasing should be considered correct as well, provided that they clearly convey the same meaning.

## Question

{{ question_text }}

## Answer key

- {{ answer_key }}

## Student answers (partly or fully incorrect)

{{ student_answers }}
''')

INTERACTIVE_QUIZ_PROMPT = Template('''You are a friendly tutor for a university course. Your name is Heymans. You are about to chat with a student about the excerpt from a textbook below. The student is a beginner, so keep questions and feedback simple.

<textbook>
{{ source }}
</textbook>

The chat session is structured as follows:

- You begin the conversation by asking the student a short, open-ended question based on the material provided above. Indicate which section of the textbook is the basis for the question.
- Evaluate the student's response to determine if it sufficiently demonstrates understanding of the concept(s).
- If the response does not connect to the question, remind the student that the assignment should be taken seriously.
- If the response resembles the textbook or your own feedback, remind the student to use his or her own words.
- If the response is satisfactory, conclude the teaching session. Do not offer to continue the conversation. End your response with <FINISHED>.
- If it the response is not satisfactory, provide constructive feedback and suggestions for improvement.
- After providing feedback, allow the student to respond with an improved answer. Continue this feedback cycle until the answer demonstrates a satisfactory understanding of the concept(s).

Remember to keep questions and feedback simple and concrete.
''')



from flask_wtf import FlaskForm
from wtforms import StringField, PasswordField, SubmitField
from wtforms.validators import DataRequired


class LoginForm(FlaskForm):
    username = StringField("Username (you don't have one)",
                           validators=[DataRequired()])
    password = PasswordField('Password', validators=[DataRequired()])
    submit = SubmitField('Log In')



import logging
import os
logger = logging.getLogger('heymans')
# Depending on whether we are running inside a Flask app or not, we need to
# instantiate the database model differently. It is difficult to determine
# automatically whether Flask is running, because at this point we're still
# in the initialization phase. Therefore, we're explicitly setting an 
# environment variable when creating the Flask app.
from sqlalchemy.orm.exc import NoResultFound
if os.environ.get('USE_FLASK_SQLALCHEMY', False):
    logger.info('using flask_sqlachemy')
    from flask_sqlalchemy import SQLAlchemy as _BaseSQLAlchemy
    db = _BaseSQLAlchemy()
    Column = db.Column
    Integer = db.Integer
    Boolean = db.Boolean
    String = db.String
    Text = db.Text
    ForeignKey = db.ForeignKey
    LargeBinary = db.LargeBinary
    DateTime = db.DateTime
    Model = db.Model
    relationship = db.relationship
    # NoResultFound = db.NoResultFound
else:
    from sqlalchemy import create_engine, Column, Integer, String, \
        ForeignKey, Text, Boolean
    from sqlalchemy.ext.declarative import declarative_base
    from sqlalchemy.orm import scoped_session, sessionmaker, relationship
    import logging
    logger.info('using standalone_sqlachemy')
    engine = create_engine('sqlite:///:memory:')
    session = scoped_session(sessionmaker(autocommit=False, autoflush=False,
                                          bind=engine))
    Base = declarative_base()
    # Create a db object that mimics Flask-SQLAlchemy's, with Model as an 
    # attribute
    class SQLAlchemy:
        def __init__(self, metadata):
            self.Model = declarative_base(metadata=metadata)
    db = SQLAlchemy(metadata=Base.metadata)
    db.session = session
    Model = db.Model
    Model.query = session.query_property()
    def init_db(): Model.metadata.create_all(bind=engine)
    def drop_db(): Model.metadata.drop_all(bind=engine)
    
    
class User(Model):
    __tablename__ = 'user'

    user_id = Column(String, primary_key=True)
    username = Column(String)

    documents = relationship('Document', back_populates='user',
                             cascade='all, delete-orphan')
    quizzes = relationship('Quiz', back_populates='user',
                           cascade='all, delete-orphan')
    interactive_quizzes = relationship(
        'InteractiveQuiz',
        back_populates='user',
        cascade='all, delete-orphan')
    interactive_quiz_conversations = relationship(
        'InteractiveQuizConversation',
        back_populates='user',
        cascade='all, delete-orphan')


class Quiz(Model):
    __tablename__ = 'quiz'

    quiz_id = Column(Integer, primary_key=True)
    user_id = Column(String, ForeignKey('user.user_id'), nullable=False)
    name = Column(String, nullable=False)
    validation = Column(Text, nullable=True)
    qualitative_error_analysis = Column(Text, nullable=True)

    # Each Quiz has multiple Questions, so we define a one-to-many relationship
    questions = relationship('Question', back_populates='quiz',
                             cascade='all, delete-orphan')

    # And is owned by a single user
    user = relationship('User', back_populates='quizzes')
    
    
class Question(Model):
    __tablename__ = 'question'

    question_id = Column(Integer, primary_key=True)
    quiz_id = Column(Integer, ForeignKey('quiz.quiz_id', ondelete='CASCADE'), nullable=False)
    name = Column(String, nullable=True)
    text = Column(Text, nullable=False)
    answer_key = Column(Text)

    # Each Question belongs to a Quiz, and has multiple Answers
    quiz = relationship('Quiz', back_populates='questions')
    attempts = relationship('Attempt', back_populates='question',
                            cascade='all, delete-orphan')


class Attempt(Model):
    __tablename__ = 'attempt'

    attempt_id = Column(Integer, primary_key=True)
    question_id = Column(Integer, ForeignKey('question.question_id'),
                         nullable=False)
    username = Column(String, nullable=False)
    answer = Column(Text, nullable=False)
    feedback = Column(Text)
    score = Column(Integer)

    # Each Answer is associated with a Question and a user
    question = relationship('Question', back_populates='attempts')


class Document(Model):
    __tablename__ = 'document'
    
    document_id = Column(Integer, primary_key=True)
    user_id = Column(String, ForeignKey('user.user_id'), nullable=False)
    public = Column(Boolean, nullable=False)
    name = Column(String, nullable=False)

    # Each Document is associated with one or more document chunks, interactive 
    # quizzes, and a user
    chunks = relationship('Chunk', back_populates='document',
                          cascade='all, delete-orphan')
    user = relationship('User', back_populates='documents')
    interactive_quizzes = relationship('InteractiveQuiz',
                                       back_populates='document',
                                       cascade='all, delete-orphan')
    

class Chunk(Model):
    __tablename__ = 'chunk'
    
    chunk_id = Column(Integer, primary_key=True)
    document_id = Column(Integer, ForeignKey('document.document_id'),
                         nullable=False)
    content = Column(Text)

    # Each Document is associated with one or more document chunks and a user
    document = relationship('Document', back_populates='chunks')


class InteractiveQuiz(Model):
    """A shareable quiz based on a single document, owned by a user."""

    __tablename__ = 'interactive_quiz'

    interactive_quiz_id = Column(Integer, primary_key=True)
    user_id = Column(String, ForeignKey('user.user_id'), nullable=False)
    document_id = Column(Integer, ForeignKey('document.document_id'),
                         nullable=False)

    # Properties
    name = Column(String, nullable=False)
    public = Column(Boolean, nullable=False, default=False)

    # Relationships
    user = relationship('User', back_populates='interactive_quizzes')
    document = relationship('Document', back_populates='interactive_quizzes')
    conversations = relationship('InteractiveQuizConversation',
                                 back_populates='interactive_quiz',
                                 cascade='all, delete-orphan')


class InteractiveQuizConversation(Model):
    """A (possibly multi-turn) chat session between a student and Heymans."""

    __tablename__ = 'interactive_quiz_conversation'

    conversation_id = Column(Integer, primary_key=True)
    interactive_quiz_id = Column(Integer,
                                 ForeignKey('interactive_quiz.interactive_quiz_id'),
                                 nullable=False)
    user_id = Column(String, ForeignKey('user.user_id'), nullable=False)

    # Properties
    finished = Column(Boolean, nullable=False, default=False)

    # Relationships
    interactive_quiz = relationship('InteractiveQuiz',
                                    back_populates='conversations')
    user = relationship('User', back_populates='interactive_quiz_conversations')
    messages = relationship('InteractiveQuizMessage',
                            back_populates='conversation',
                            cascade='all, delete-orphan')


class InteractiveQuizMessage(Model):
    """An individual message (question, answer, system prompt, etc.)
    in an InteractiveQuizConversation.
    """

    __tablename__ = 'interactive_quiz_message'

    message_id = Column(Integer, primary_key=True)
    conversation_id = Column(Integer,
                             ForeignKey('interactive_quiz_conversation.conversation_id'),
                             nullable=False)

    # Properties
    text = Column(Text, nullable=False)
    message_type = Column(String, nullable=False)  # e.g. 'user' or 'ai'

    # Relationships
    conversation = relationship('InteractiveQuizConversation',
                                back_populates='messages')


import os
import base64

# SERVER
#
# The external server address, that is, the URL that users visit
server_url = os.environ.get('FLASK_SERVER_URL', 'http://127.0.0.1:5000')
# The port at which the flask server is running internally. This can be
# different from the server URL if the app is running behind a proxy that
# redirects
flask_port = int(os.environ.get('FLASK_PORT', 5000))
# The flask host arhument where all 0s means listen to all incoming addresses
flask_host = os.environ.get('FLASK_HOST', '0.0.0.0')
# The secret key is used for logging in. This should be a long and arbitrary
# string that is hard to guess. This should not be shared
flask_secret_key = os.environ.get('FLASK_SECRET_KEY', '0123456789ABCDEF')

encryption_salt = base64.urlsafe_b64decode(
    os.environ.get("ENCRYPTION_SALT", "ZGVmYXVsdF9zYWx0X2Zvcl90ZXN0aW5n")
)
# GOOGLE SSO OPTIONS:
google_login_enabled = True
google_client_id = os.environ.get("GOOGLE_CLIENT_ID", None)
google_client_secret = os.environ.get("GOOGLE_CLIENT_SECRET", None)
google_discovery_url = (
    "https://accounts.google.com/.well-known/openid-configuration"
)
google_redirect_uri = os.environ.get("GOOGLE_REDIRECT_URI", None)
# Default model to use
default_model = os.environ.get('HEYMANS_DEFAULT_MODEL', 'claude-4-5-sonnet')

# DEV OPTIONS
#
# Use a dummy model during development
dummy_model = bool(int(os.environ.get('HEYMANS_DUMMY_MODEL', 0)))
dummy_delay = .1

# QUIZ GRADING
#
# The minimum number of characters that an answer should have
min_answer_length = 2
# The maximum duration in seconds that grading a single batch should take
grading_task_timeout = 300
validation_task_timeout = 300
grading_task_max_concurrent = 50
grading_max_retries = 5

# DOCUMENTS
#
# The maximum chunk size in characters
document_max_chunk_size = 10000

# DEBUG INFO
show_debug_info = bool(os.environ.get('HEYMANS_SHOW_DEBUG_INFO', 0))



import time
from sigmund.model import model as sigmund_model
from . import config


class Dummy:
    def __init__(self, reply='dummy reply', delay=0):
        self.reply = reply
        self.delay = delay

    def predict(self, *args, **kwargs):
        time.sleep(self.delay)
        return self.reply


def chatbot_model(model: str, dummy_reply: str = 'dummy reply') -> object:
    """Initializes a Sigmund chatbot model, or a dummy model if we're running
    in dummy mode. 
    """
    if config.dummy_model:
        return Dummy(reply=dummy_reply, delay=config.dummy_delay)
    return sigmund_model(None, model)



import logging
import json
from ..models import db, Quiz, Question, Attempt, User
from ..schemas import QuizSchema

logger = logging.getLogger('heymans')


def get_or_create(session, model, **kwargs):
    """Get or create a model instance while avoiding duplicate creation."""
    instance = session.query(model).filter_by(**kwargs).one_or_none()
    if instance:
        return instance
    instance = model(**kwargs)
    session.add(instance)
    return instance


def _get_quiz(quiz_id: int, user_id: int):
    """Return the Quiz instance or raise ValueError if it doesn’t exist / is
    not owned by *user_id*.  NOTE: caller is expected to be inside a session
    scope already.
    """
    logger.info('getting quiz %s for user %s', quiz_id, user_id)
    quiz = (
        db.session.query(Quiz)
        .filter_by(quiz_id=quiz_id, user_id=user_id)
        .one_or_none()
    )
    if quiz is None:
        raise ValueError(f'no quiz {quiz_id} found for user {user_id}')
    return quiz


# ---------------------------------------------------------------------------
# CRUD helpers
# ---------------------------------------------------------------------------

def _quiz_name_exists(name: str, user_id: int,
                      exclude_quiz_id: int | None = None) -> bool:
    """Return True if a quiz with the given name exists for the user.

    Parameters
    ----------
    name : str
        The quiz name to check.
    user_id : int
        The owner/user id to scope uniqueness.
    exclude_quiz_id : int or None, optional
        If provided, exclude this quiz id from the existence check.
    """
    query = db.session.query(Quiz.quiz_id).filter(
        Quiz.user_id == user_id,
        Quiz.name == name,
    )
    if exclude_quiz_id is not None:
        query = query.filter(Quiz.quiz_id != exclude_quiz_id)
    return db.session.query(query.exists()).scalar()


def _unique_quiz_name(base_name: str, user_id: int,
                      exclude_quiz_id: int | None = None) -> str:
    """Compute a unique quiz name for a user by appending numeric suffixes.

    Parameters
    ----------
    base_name : str
        Desired base name.
    user_id : int
        The owner/user id to scope uniqueness.
    exclude_quiz_id : int or None, optional
        If provided, exclude this quiz id from the uniqueness check (useful for updates).

    Returns
    -------
    str
        A unique name scoped to the user.
    """
    base = base_name
    candidate = base
    counter = 1
    while _quiz_name_exists(candidate, user_id,
                            exclude_quiz_id=exclude_quiz_id):
        candidate = f"{base} ({counter})"
        counter += 1
    return candidate


def new_quiz(name: str, user_id: int) -> int:
    """Create a new empty quiz with a unique name for the given user.

    Ensures uniqueness by appending a numeric suffix like " (1)", " (2)", etc.,
    when a quiz with the same name already exists for the same user.

    Parameters
    ----------
    name : str
        Desired quiz name.
    user_id : int
        The ID of the user who owns the quiz.

    Returns
    -------
    int
        The primary key (quiz_id) of the newly created quiz.

    Notes
    -----
    - Uniqueness is enforced per user only.
    - All database operations occur within a transaction block.
    """
    with db.session.begin():
        unique_name = _unique_quiz_name(name, user_id)
        quiz = Quiz(name=unique_name, user_id=user_id)
        db.session.add(quiz)
        db.session.flush()
        return quiz.quiz_id


def update_quiz(quiz_id: int, quiz_info: dict, user_id: int) -> None:
    """Replace an existing quiz’s metadata, questions, and attempts.

    Ensures the quiz name remains unique per user. When updating, if the
    requested name equals the current quiz name, it will not be treated as a duplicate.
    Otherwise, a numeric suffix like " (1)", " (2)", etc., is appended until unique.

    Parameters
    ----------
    quiz_id : int
        The primary key of the quiz to update.
    quiz_info : dict
        Structure with updated quiz metadata, questions, and attempts.
    user_id : int
        Owner of the quiz, used to scope uniqueness and authorization.

    Notes
    -----
    - All operations occur within a transaction.
    - Existing questions and attempts are fully replaced.
    """
    with db.session.begin():
        quiz = _get_quiz(quiz_id, user_id)

        # Quiz-level fields
        if 'name' in quiz_info:
            requested_name = quiz_info['name']
            # Only compute a unique name if it differs or conflicts with others
            if requested_name != quiz.name:
                quiz.name = _unique_quiz_name(requested_name, user_id,
                                              exclude_quiz_id=quiz_id)
        if 'validation' in quiz_info:
            quiz.validation = quiz_info['validation']
        else:
            quiz.validation = None

        if 'qualitative_error_analysis' in quiz_info:
            quiz.qualitative_error_analysis = quiz_info['qualitative_error_analysis']
        else:
            quiz.qualitative_error_analysis = None

        # Remove existing questions (cascade deletes attempts if configured)
        for question in list(quiz.questions):
            db.session.delete(question)

        # Add new questions & attempts
        for q_info in quiz_info['questions']:
            question = Question(
                text=q_info['text'],
                name=q_info['name'],
                answer_key=json.dumps(q_info['answer_key']),
                quiz=quiz,
            )
            db.session.add(question)
            for a_info in q_info.get('attempts', []):
                attempt = Attempt(
                    answer=a_info['answer'],
                    question=question,
                    username=a_info['username'],
                )
                db.session.add(attempt)


def get_quiz(quiz_id: int, user_id: int) -> dict:
    """Return full quiz information (questions + attempts) as a dict."""
    with db.session.begin():
        quiz_dict = QuizSchema().dump(_get_quiz(quiz_id, user_id))

    # Unpack JSON columns outside the TX scope (pure Python work)
    for question in quiz_dict['questions']:
        question['answer_key'] = json.loads(question['answer_key'])
        for attempt in question.get('attempts', []):
            if attempt['feedback'] is not None:
                attempt['feedback'] = json.loads(attempt['feedback'])
    return quiz_dict


def delete_quiz(quiz_id: int, user_id: int) -> None:
    """Delete a quiz (and cascaded children)."""
    with db.session.begin():
        db.session.delete(_get_quiz(quiz_id, user_id))


def list_quizzes(user_id: int) -> list:
    """Return ``[{quiz_id, name}, …]`` for all quizzes owned by *user_id*."""
    with db.session.begin():
        all_quizzes = db.session.query(Quiz).filter_by(user_id=user_id).all()
        quizzes = QuizSchema(many=True, only=['quiz_id', 'name']).dump(all_quizzes)
    return quizzes


def update_attempts(quiz_id: int, quiz_data: dict, user_id: int) -> None:
    """Bulk-update attempts based on a list of attempt dicts."""
    attempts = []
    for question in quiz_data['questions']:
        attempts += question['attempts']
    with db.session.begin():
        # First commit the qualitative error analysis
        quiz = _get_quiz(quiz_id, user_id)
        quiz.qualitative_error_analysis = \
            quiz_data.get('qualitative_error_analysis', '')
        # Then commit all attempts
        for a_dict in attempts:
            attempt = db.session.get(Attempt, a_dict['attempt_id'])
            if not attempt:
                logger.warning('Missing attempt: %s', a_dict['attempt_id'])
                continue

            quiz = attempt.question.quiz if attempt.question else None
            if not quiz or quiz.user_id != user_id:
                logger.warning(
                    f'User {user_id} not allowed to edit attempt owned by {quiz.user_id}')
                continue

            attempt.score = a_dict['score']
            if a_dict['feedback'] is not None:
                attempt.feedback = json.dumps(a_dict['feedback'])



import json
from flask import Blueprint, request, jsonify
from flask_login import login_required, current_user
import logging
from . import missing_file, error, success, not_found
from ..database.operations import documents as ops

logger = logging.getLogger('heymans')
documents_api_blueprint = Blueprint('api/documents', __name__)


@documents_api_blueprint.route('/add', methods=['POST'])
@login_required
def add():
    """Add a new document with an associated file. This expects multipart form
    data consistting of a single file and JSON data. If no name is specified, 
    the file name is used.

    Request JSON Example
    --------------------
    {
        "public": <bool>,
        "name": <str>    # optional
    }
    
    Reply JSON Example
    ------------------
    {
        "document_id": <int>,
        "chunk_ids": [<int>, ...]
    }

    Returns
    -------
    200 OK
    400 Bad Request
    """
    document_info = json.loads(request.form.get('json', ''))
    if 'file' not in request.files:
        return missing_file()
    file = request.files['file']
    public = document_info.get('public')
    filename = file.filename
    file_content = file.read()
    mimetype = file.content_type
    name = document_info.get('name', filename)
    try:
        document_id, chunk_ids = ops.add_document(
            current_user.get_id(), public, name, file_content, filename,
            mimetype)
    except Exception as e:
        logger.error(f"Error adding document: {str(e)}")
        return error(str(e))
    logger.info(f'Added document: {document_id}')
    return jsonify({'document_id': document_id, 'chunk_ids': chunk_ids})


@documents_api_blueprint.route('/update/<int:document_id>', methods=['POST'])
@login_required
def update(document_id):
    """Update the public status and/ or name of an existing document.

    Request JSON example
    --------------------
    {
        "public": <bool>,    # optional
        "name": <str>        # optional
    }

    Returns
    -------
    200 OK
    400 Bad Request
    404 Not Found
    """
    public = request.json.get('public')
    name = request.json.get('name')
    if ops.update_document(current_user.get_id(), document_id, public, name):
        return success()
    return not_found('document does not exist or belongs to different user')


@documents_api_blueprint.route('/delete/<int:document_id>', methods=['DELETE'])
@login_required
def delete(document_id):
    """Delete an existing document by its identifier.

    Returns
    -------
    200 OK
    404 Not Found
    """
    if ops.delete_document(current_user.get_id(), document_id):
        return success()
    return not_found('document does not exist or belongs to different user')


@documents_api_blueprint.route('/list/<int:include_public>')
@login_required
def list_(include_public):
    """List all documents belonging to the user, optionally including public
    documents from other users.

    Path Parameters
    ---------------
    include_public : int
        0 to return only user's own documents;
        1 to include all public documents from other users.
        
    Reply JSON Example
    ------------------
    [
        {
            "document_id": <int>,
            "name": <str>,
            "public": <bool>,
            ...
        },
        ...
    ]

    Returns
    -------
    200 OK
    """
    return jsonify(ops.list_documents(current_user.get_id(), include_public))
